// Vincia.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2017 P.Z. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains global header information for Vincia.

#ifndef VINCIA_H
#define VINCIA_H

// Include Pythia 8 headers 
#include "Pythia8/Pythia.h"

// Global VINCIA functionality
#include "Vincia/VinciaCommon.h"

// Now include headers that need this information
#include "Vincia/Reconnections.h"
#include "Vincia/SubtractionTerms.h"
#include "Vincia/TimeAntShower.h"
#include "Vincia/SpaceAntShower.h"
#include "Vincia/MG4interface.h"
#include "Vincia/MG5interface.h"
#include "Vincia/MHV.h"
#include "Vincia/Merging.h"
#include "Vincia/MergingHooks.h"

// Define namespace inside which Vincia lives
namespace Vincia { 
  
//==========================================================================
  
// The VinciaPlugin class
  
class VinciaPlugin {
    
 public:
    
  VinciaPlugin() {
    cerr << "\n (VinciaPlugin():) cannot construct VinciaPlugin :(\n\n"
         <<"  Sorry, the Vincia plugin must be constructed with a pointer"
         <<" to an\n  explicit Pythia object (since it's a plugin)." 
         <<" Please use one of \n\n"
         <<"     VinciaPlugin vincia(Pythia* pythia);\n"
         <<"     VinciaPlugin vincia(Pythia* pythia, string xmlDir);\n"
         <<"     VinciaPlugin vincia(Pythia* pythia, string xmlDir, string xmlIndexFile);\n"
         <<"\n";
  }
    
  // Constructor
  VinciaPlugin(Pythia* pythia, 
    string xmlDir="../share/Vincia/xmldoc", 
    string xmlIndexFile="VinciaIndex.xml") {      
      
    // First set pointers to Pythia classes
    pythiaPtr         = pythia;
    couplingsPtr      = &pythia->couplings;
    infoPtr           = &pythia->info;
    particleDataPtr   = &pythia->particleData;
    partonSystemsPtr  = &pythia->partonSystems;
    rndmPtr           = &pythia->rndm;
    settingsPtr       = &pythia->settings;
    // Initial value for pointer to user hooks.
    userHooksPtr      = 0;
      
    // Adding normalization factors here to not have to change
    // the code
    settingsPtr->addParm("Vincia:pTnormalization",4.0,true,true,1.0,4.0);   
    settingsPtr->addParm("Vincia:mDnormalization",2.0,true,true,2.0,2.0);
    settingsPtr->addParm("Vincia:pTnormalizationII",1.0,true,true,1.0,1.0);
    settingsPtr->addParm("Vincia:pTnormalizationIF",1.0,true,true,1.0,1.0);
    settingsPtr->addParm("Vincia:mSnormalizationI",1.0,true,true,1.0,1.0);
    settingsPtr->addParm("Vincia:mCnormalizationI",1.0,true,true,1.0,1.0);
    settingsPtr->addParm("Vincia:mSnormalizationF",1.0,true,true,1.0,1.0);
      
    // Find path to PYTHIA data files. 
    const char* PYTHIA8DATA = "PYTHIA8DATA";
    char* envPath = getenv(PYTHIA8DATA);
    if (envPath == 0 || *envPath == '\0') {
      cout<<" VINCIA Error: you have not set the PYTHIA8DATA environment variable.\n";
      cout<<" VINCIA Hint: if your local PYTHIA install is in the directory $HOME/yourpythia, try setting PYTHIA8DATA to $HOME/yourpythia/share/Pythia8/xmldoc"<<endl;
      cout<<"              alternatively, for a default global PYTHIA install, use /usr/share/Pythia8/xmldoc"<<endl; 
      return;
    }
      
    // Find path to VINCIA data files, i.e. xmldoc directory location.
    // Environment variable takes precedence, else use constructor input.
    string xmlPath        = "";
    bool existsVINCIADATA = false;
    const char* VINCIADATA = "VINCIADATA";
    envPath = getenv(VINCIADATA);
    if (envPath != 0 && *envPath != '\0') {
      int i = 0;
      while (*(envPath+i) != '\0') xmlPath += *(envPath+(i++));
      existsVINCIADATA = true;
    }
    else xmlPath = xmlDir;
    if (xmlPath[ xmlPath.length() - 1 ] != '/') xmlPath += "/";      
    // Set the VINCIADATA environment variable, for use in main programs
    setenv(VINCIADATA,xmlPath.c_str(),0);
    // Save the presumed location of VINCIA runtime data files
    VINCIA_SHARE = xmlPath+"../";
    // Add the VINCIA list of parameters to the PYTHIA 8 settings database
    settingsPtr->init(xmlPath+xmlIndexFile, true); 

    // Verbose level
    setVerbose(settingsPtr->mode("Vincia:verbose"));      

    // Init FSR shower pointers and default settings
    // ( beyond those set by the non-virtual TimeShower::initPtr() )
    fsrShower.functionsPtr           = &functions;
    fsrShower.resolutionPtr          = &resolution;
    fsrShower.vinCluPtr              = &vinClu;
    fsrShower.vinComPtr              = &vinCom;
    fsrShower.reconnectionsPtr       = &reconnections;
    fsrShower.ISRshowerPtr           = &isrShower;
    fsrShower.matrixElementCorrectionsPtr = &matrixElementCorrections;
    fsrShower.vinMergingHooksPtr     = &vinMergingHooks;
    fsrShower.weightsPtr             = &vinWeights;
      
    // Init ISR shower pointers and default settings
    // ( beyond those set by the non-virtual SpaceShower::initPtr() )
    isrShower.functionsPtr           = &functions;
    isrShower.resolutionPtr          = &resolution;
    isrShower.vinCluPtr              = &vinClu;
    isrShower.vinComPtr              = &vinCom;
    isrShower.reconnectionsPtr       = &reconnections;
    isrShower.FSRshowerPtr           = &fsrShower;
    isrShower.matrixElementCorrectionsPtr = &matrixElementCorrections;
    isrShower.vinMergingHooksPtr     = &vinMergingHooks;
    isrShower.weightsPtr             = &vinWeights;
 
    // Check for DATA directory
    if (verbose >= 2) {
      if (existsVINCIADATA) 
        cout<<" VinciaPlugin(): VINCIADATA environment variable ok."<<endl;
      else cout<<" VinciaPlugin(): VINCIADATA environment variable undefined. Using constructor instead."<<endl; 
    }
    if (verbose >= 1) {
      cout<<" VinciaPlugin(): using configuration directory : "
          <<xmlPath<<endl;
    }
      
    // Set Pythia 8 shower pointers to Vincia
    pythiaPtr->setShowerPtr(&fsrShower, &fsrShower, &isrShower);

    // Set the hook for resonance scales
    pythiaPtr->setUserHooksPtr( &resScaleHook );
    
  }

  // Initialize Vincia & Pythia
  void init(string commandFile="none", int subrun=-999) {

    // Get verbose level
    setVerbose(settingsPtr->mode("Vincia:verbose"));
    
    // Verbose output
    if (verbose >= 6)
      cout<<" VinciaPlugin()::init() begin ------------------"<<endl;      
      
    // Vincia on/off
    vinciaOn    = settingsPtr->flag("Vincia");
      
    // Initialise VINCIA settings
    if (vinciaOn) {

      // Get list of Pythia readString commands given to Settings so far      
      vector<string> readStringHistory = settingsPtr->getReadHistory();

      // Reset tune parameters using default (VINCIA-specific) tune       
      string defaultTune = VINCIA_SHARE + "/tunes/default.cmnd";
      // Check if this file exists, otherwise look in VINCIA_SHARE
      if (fileExists(defaultTune)) {
        if (verbose >= 3) 
          cout << " VinciaPlugin(): reading default TUNE preset from "
               <<defaultTune<<endl;
        pythiaPtr->readFile(defaultTune);
      }

      // Then reapply readString commands to enable user modifications
      // (but ignore any Tune:ee or Tune:pp commands)
      if (verbose >= 3) cout<<" VinciaPlugin(): reapplying user settings"<<endl;
      for (int i=0; i<(int)readStringHistory.size(); ++i) {
        string lineNow = toLower(readStringHistory[i]);
        if (lineNow.find("tune:ee") == string::npos && lineNow.find("tune:pp")
          == string::npos) pythiaPtr->readString(lineNow);
      }
      
      // Read any command file given to init 
      if (commandFile != "none") {
        if (verbose >= 3) cout<<" VinciaPlugin(): processing "<<commandFile<<endl;
        if (subrun >= 0) pythiaPtr->readFile(commandFile,subrun); 
        else pythiaPtr->readFile(commandFile);
      }
      
      // Check if verbose setting changed
      verbose = settingsPtr->mode("Vincia:verbose");
      
      if (verbose >= 3)
        cout<<" VinciaPlugin(): initializing antenna sets..."<<endl;

      // FSR antenna sets
      antennaSetFSR.initPtr(&functions, particleDataPtr, 
        rndmPtr, settingsPtr);
      antennaSetFSR.init();
      // ISR antenna sets
      antennaSetISR.initPtr(&functions,  particleDataPtr, 
        rndmPtr, settingsPtr);
      antennaSetISR.init();
        
      if (verbose >= 3)
        cout<<" VinciaPlugin(): initializing pointers..."<<endl;
        
      // Hand antenna set pointers to shower and matching objects
      fsrShower.initAntPtr(&antennaSetFSR);
      isrShower.initAntPtr(&antennaSetISR);
      matrixElementCorrections.initAntPtr(&antennaSetFSR,&antennaSetISR);
    } 
      
    // If VINCIA off, let us know about it
    else {

      // Set Pythia 8 shower pointers back to Pythia internal ones
      pythiaPtr->setShowerPtr(0, 0, 0);

      if (verbose >= 2) printOut("VinciaPlugin::init()",
        "Vincia is switched off");
    }

    // Pass pointers on to objects that require them
    resolution.initPtr(settingsPtr);
    rambo.initPtr(rndmPtr);
    subtractionTerms.initPtr(settingsPtr);
    vinCom.initPtr(particleDataPtr,settingsPtr);
    vinClu.initPtr(rndmPtr,settingsPtr,&vinCom);
    mg4interface.initPtr(couplingsPtr,particleDataPtr,settingsPtr,&vinCom);
    SusyLesHouches* slhaPtr = &pythiaPtr->slhaInterface.slha;
    mg5interface.initPtr(couplingsPtr,particleDataPtr,rndmPtr,settingsPtr,
      slhaPtr,&vinCom);
    mhvAmplitudes.initPtr(couplingsPtr, particleDataPtr, settingsPtr);
    matrixElementCorrections.initPtr(settingsPtr, infoPtr, rndmPtr,
      &mg4interface, &mg5interface, &mhvAmplitudes, &vinCom);
    reconnections.initPtr(rndmPtr,settingsPtr);
    vinWeights.initPtr(settingsPtr, infoPtr, &vinCom);

    // Pass pointers on to merging
    vinMergingHooks.setVinciaPtrs(&fsrShower, &isrShower, &vinCom);
    vinMerging.setVinciaPtrs(&vinMergingHooks, &fsrShower, &isrShower, &vinCom);

    // Pass merging pointers to Pythia
    pythiaPtr->setMergingPtr(&vinMerging);
    pythiaPtr->setMergingHooksPtr(&vinMergingHooks);

    // Initialise 
    vinCom.init();
    resolution.init();
    reconnections.init();
    vinClu.init();
    vinWeights.init();

    // Initialise Pythia
    pythiaPtr->init();

    // MECs depend on Pythia Couplings so must be initialised after Pythia
    matrixElementCorrections.init();
    subtractionTerms.init();

  }

  //*********
    
  // Automatically set verbose level in all inheriting members
  void setVerbose(int verboseIn) {
    verbose = verboseIn;
    vinCom.setVerbose(verboseIn);
    resolution.setVerbose(verboseIn);
    fsrShower.setVerbose(verboseIn);
    isrShower.setVerbose(verboseIn);
    reconnections.setVerbose(verboseIn);
    mg4interface.setVerbose(verboseIn);
    mg5interface.setVerbose(verboseIn);
    vinClu.setVerbose(verboseIn);
    mhvAmplitudes.setVerbose(verboseIn);
    matrixElementCorrections.setVerbose(verboseIn);
  }
    
  // Main VINCIA on/off switch
  bool vinciaOn;
    
  // Instantiate Vincia Objects
  VinciaCommon        vinCom;
  Resolution          resolution;
  TimeAntShower       fsrShower;
  SpaceAntShower      isrShower;
  Reconnections       reconnections;
  SubtractionTerms    subtractionTerms;
  ResScaleHook        resScaleHook;
  VinciaWeights       vinWeights;
  MatrixElementCorrections matrixElementCorrections;
  VinciaMerging       vinMerging;
  VinciaMergingHooks  vinMergingHooks;
  MHVamplitudes       mhvAmplitudes;

  // Auxiliary objects
  Functions           functions;
  MG4interface        mg4interface;
  MG5interface        mg5interface;
  Rambo               rambo;
  VinClu              vinClu;  
    
  // Vectors of antenna functions
  AntennaSetFSR antennaSetFSR;
  AntennaSetISR antennaSetISR;

  // Location of runtime VINCIA files
  string VINCIA_SHARE;

  // Pointers to Pythia classes
  Pythia*        pythiaPtr;
  Couplings*     couplingsPtr;
  Info*          infoPtr;
  ParticleData*  particleDataPtr;
  PartonSystems* partonSystemsPtr;
  Rndm*          rndmPtr; 
  Settings*      settingsPtr; 
  UserHooks*     userHooksPtr;

  // Shortcut for post-Vincia event weight
  double weight(int iVar=0) {
    if (settingsPtr->flag("Vincia"))
      return vinWeights.getWeight(iVar);
    return infoPtr->weight();
  }
  string weightLabel(int iVar=0) {
    if (settingsPtr->flag("Vincia"))
      return vinWeights.getWeightLabel(iVar);
    return "";
  }
  int  nWeights()         { return vinWeights.nWeights(); }
  bool reweightingOccurred() { return vinWeights.reweightingOccurred(); }
    
  // Utilities for printing Info and Internal histograms
  void printInfo() { 
    fsrShower.printInfo(true);
    isrShower.printInfo(true);
  }
  void printHistos() { fsrShower.printHistos(); }
  void writeHistos( string fileName = "vincia", string lastName = "dat" ) { 
    fsrShower.writeHistos(fileName,lastName); 
  }
  const Hist& getDiagnosticHistogram( string name ) {
    return fsrShower.getDiagnosticHistogram(name);
  }

 private :

  // Verbose level
  int verbose;

}; // end Vincia plugin
  
} // end Vincia namespace
#endif
