// Merging.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator. 
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2017 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

#include "Vincia/Merging.h"

namespace Vincia {

  const bool debugNow = false;

  //==========================================================================

  // The Merging class.

  // Factor by which the maximal value of the merging scale can deviate before
  // a warning is printed.
  const double VinciaMerging::TMSMISMATCH = 1.5;

  //--------------------------------------------------------------------------

  // Initialize.

  void VinciaMerging::init() {
    if (debugNow) cout << " VinciaMerging::init begin" << endl;

    // Reset minimal tms value.
    tmsNowMin = infoPtr->eCM();

    if (debugNow) cout << " VinciaMerging::init end" << endl;
  }

  //--------------------------------------------------------------------------

  // Main function to perform the merging.

  int VinciaMerging::mergeProcess(Event& process) {

    if (debugNow) cout << " VinciaMerging::mergeProcess begin" << endl;

    // Reinitialise hard process.
    vcMergingHooksPtr->hardProcess->clear();
    vcMergingHooksPtr->processSave = settingsPtr->word("Merging:Process");
    vcMergingHooksPtr->hardProcess->initOnProcess(
      settingsPtr->word("Merging:Process"), particleDataPtr);

    vcMergingHooksPtr->nRequestedSave = 
      settingsPtr->mode("Merging:nRequested");

    // Always initialize with unordered and incomplete path.
    // To be reset by chosen history.
    vcMergingHooksPtr->useOrderedPath(false);
    vcMergingHooksPtr->useCompletePath(false);
    vcMergingHooksPtr->doOrderedMECs(false);

    // Ensure that merging weight is not counted twice.
    bool includeWGT = vcMergingHooksPtr->includeWGTinXSEC();

    // Possibility to apply merging scale to an input event.
    bool applyTMSCut = settingsPtr->flag("Merging:doXSectionEstimate");
    if (applyTMSCut) {
      if (cutOnProcess(process)) {
        if (includeWGT) infoPtr->updateWeight(0.0);
        return -1;
      } else return 1;
    }

    // Ensure that merging hooks to not veto events in the trial showers.
    vcMergingHooksPtr->doIgnoreStep(true);
    // For now, prefer construction of ordered histories.
    vcMergingHooksPtr->orderHistories(true);
    // For pp > h, allow cut on state, so that underlying processes
    // can be clustered to gg > h
    if ( vcMergingHooksPtr->getProcessString().compare("pp>h") == 0 )
      vcMergingHooksPtr->allowCutOnRecState(true);

    // Reset weight of the event.
    double wgt = 1.0;
    vcMergingHooksPtr->setWeightCKKWL(1.0);
    vcMergingHooksPtr->muMI(-1.0);

    // Prepare process record for merging. If Pythia has already decayed
    // resonances used to define the hard process, remove resonance decay
    // products.
    Event newProcess( vcMergingHooksPtr->bareEvent(process,true) );
    // Store candidates for the splitting V -> qqbar'
    vcMergingHooksPtr->storeHardProcessCandidates(newProcess);

    // Check if event passes the merging scale cut.
    double tmsval = vcMergingHooksPtr->tms();
    // Get merging scale in current event.
    double tmsnow = vcMergingHooksPtr->tmsCalc(newProcess);
    // Calculate number of clustering steps
    int nSteps = vcMergingHooksPtr->getNumberOfClusteringSteps(newProcess,true);

    // Check if hard event cut should be applied later.
    bool allowReject = settingsPtr->flag("Merging:applyVeto");

    // Too few steps can be possible if a chain of resonance decays has been
    // removed. In this case, reject this event, since it will be handled in
    // lower-multiplicity samples.
    int nRequested = vcMergingHooksPtr->nRequested();

    if (nSteps < nRequested && allowReject) {
      if (includeWGT) infoPtr->updateWeight(0.0);
      else vcMergingHooksPtr->setWeightCKKWL(0.0);
      return -1;
    }

    // Reset the minimal tms value, if necessary.
    tmsNowMin = (nSteps == 0) ? 0. : min(tmsNowMin, tmsnow);

    // Get random number to choose a path.
    double RN = rndmPtr->flat();

    // Set dummy process scale.
    newProcess.scale(0.0);
    // Generate all histories.
    VinciaHistory FullHistory( nSteps, 0.0, newProcess, VinciaClustering(),
      vcMergingHooksPtr, (*beamAPtr), (*beamBPtr), particleDataPtr, infoPtr,
      timesAntShowerPtr, spaceAntShowerPtr, coupSMPtr, true, true, true, true,
      1.0, 0);

    // Project histories onto desired branches, e.g. only ordered paths.
    FullHistory.projectOntoDesiredHistories();

    // Set the number of good branches.
    vcMergingHooksPtr->nrGoodBranch(FullHistory.goodBranches.size());

    // Setup the selected path.
    FullHistory.select(RN)->setSelectedChild();

    // Do not apply cut if the configuration could not be projected onto an
    // underlying born configuration.
    bool applyCut = allowReject && nSteps > 0 && 
      FullHistory.select(RN)->nClusterings() > 0;

    // Enfore merging scale cut if the event did not pass the merging scale
    // criterion.
    bool enforceCutOnLHE = settingsPtr->flag("Merging:enforceCutOnLHE");
    if ( enforceCutOnLHE && applyCut && tmsnow < tmsval ) {
      infoPtr->errorMsg("Warning in VinciaMerging::mergeProcess: Les Houches"
        " Event fails merging scale cut. Reject event.");
      if (includeWGT) infoPtr->updateWeight(0.0);
      else vcMergingHooksPtr->setWeightCKKWL(0.0);
      return -1;
    }

    bool complete = (FullHistory.select(RN)->nClusterings() == nSteps);

    if (!complete)
      infoPtr->errorMsg("Warning in VinciaMerging::mergeProcess: No clusterings"
        " found. History incomplete.");
  
    // Check if chosen path is ordered, and reset if necessary.
    double scaleMax    = FullHistory.select(RN)->hardFacScale(
      FullHistory.select(RN)->state);
    bool isOrderedPath = FullHistory.select(RN)->isOrderedPath(scaleMax);
    // Set ordering and completeness.
    vcMergingHooksPtr->useOrderedPath(isOrderedPath);
    vcMergingHooksPtr->useCompletePath(complete);
    scaleMax      = infoPtr->eCM();
    isOrderedPath = FullHistory.select(RN)->isOrderedPath(scaleMax);
    // No ordered MECs in trial shower if no ordered and complete path.
    vcMergingHooksPtr->doOrderedMECs(isOrderedPath && complete);

    // Calculate CKKWL weight:
    // Perform reweighting with Sudakov factors, save alphaS ratio and
    // PDF ratio weights.
    wgt = FullHistory.weightTREE( trialPartonLevelPtr, &vinComPtr->alphaS,
      &vinComPtr->alphaS, &vinComPtr->alphaEM, &vinComPtr->alphaEM, RN);

    // Now ordered MECs are allowed again.
    vcMergingHooksPtr->doOrderedMECs(true);
 
    // Event with production scales set for further (trial) showering
    // and starting conditions for the shower.
    FullHistory.getStartingConditions( RN, process );
    // If necessary, reattach resonance decay products.
    vcMergingHooksPtr->reattachResonanceDecays(process);

    // Allow to dampen histories in which the lowest multiplicity reclustered
    // state does not pass the lowest multiplicity cut of the matrix element.
    double dampWeight = vcMergingHooksPtr->dampenIfFailCuts(
      FullHistory.lowestMultProc(RN));
    // Save the weight of the event for histogramming. Only change the
    // event weight after trial shower on the matrix element
    // multiplicity event (= in doVetoStep).
    wgt *= dampWeight;

    // Save the weight of the event for histogramming.
    if (!includeWGT) vcMergingHooksPtr->setWeightCKKWL(wgt);

    // Update the event weight.
    double norm = (abs(infoPtr->lhaStrategy()) == 4) ? 1/1e9 : 1.0;
    if (includeWGT) infoPtr->updateWeight(infoPtr->weight()*wgt*norm);

    // Allow merging hooks to veto events from now on.
    vcMergingHooksPtr->doIgnoreStep(false);

    // If no-emission probability is zero.
    if (allowReject && wgt == 0.0) return 0;

    // Done.
    return 1;

  }

  //--------------------------------------------------------------------------

  // Function to apply the merging scale cut on an input event.
  bool VinciaMerging::cutOnProcess(Event& process) {

    if (debugNow) cout << " VinciaMerging::cutOnProcess begin" << endl;

    // Save number of looping steps
    vcMergingHooksPtr->nReclusterSave = settingsPtr->mode("Merging:nRecluster");

    // For now, prefer construction of ordered histories.
    vcMergingHooksPtr->orderHistories(true);
    // For pp > h, allow cut on state, so that underlying processes
    // can be clustered to gg > h
    if ( vcMergingHooksPtr->getProcessString().compare("pp>h") == 0)
      vcMergingHooksPtr->allowCutOnRecState(true);

    // Prepare process record for merging. If Pythia has already decayed
    // resonances used to define the hard process, remove resonance decay
    // products.
    Event newProcess( vcMergingHooksPtr->bareEvent(process,true) );
    // Store candidates for the splitting V -> qqbar'
    vcMergingHooksPtr->storeHardProcessCandidates(newProcess);

    // Check if event passes the merging scale cut.
    double tmsval = vcMergingHooksPtr->tms();
    // Get merging scale in current event.
    double tmsnow = vcMergingHooksPtr->tmsCalc(newProcess);
    // Calculate number of clustering steps
    int nSteps = vcMergingHooksPtr->getNumberOfClusteringSteps(newProcess,true);

    // Too few steps can be possible if a chain of resonance decays has been
    // removed. In this case, reject this event, since it will be handled in
    // lower-multiplicity samples.
    int nRequested = vcMergingHooksPtr->nRequested();
    if (nSteps < nRequested) return true;

    // Reset the minimal tms value, if necessary.
    tmsNowMin = (nSteps == 0) ? 0. : min(tmsNowMin, tmsnow);

    // Get random number to choose a path.
    double RN = rndmPtr->flat();
    // Set dummy process scale.
    newProcess.scale(0.0);
    // Generate all histories
    VinciaHistory FullHistory( nSteps, 0.0, newProcess, VinciaClustering(),
      vcMergingHooksPtr, (*beamAPtr), (*beamBPtr), particleDataPtr, infoPtr,
      timesAntShowerPtr, spaceAntShowerPtr, coupSMPtr, true, true, true, true,
      1.0, 0);
    // Project histories onto desired branches, e.g. only ordered paths.
    FullHistory.projectOntoDesiredHistories();

    // Cut if no history passes the cut on the lowest-multiplicity state.
    double dampWeight = vcMergingHooksPtr->dampenIfFailCuts(
      FullHistory.lowestMultProc(RN));
    if (dampWeight == 0.0) return true;

    // Do not apply cut if the configuration could not be projected onto an
    // underlying born configuration.
    if ( nSteps > 0 && FullHistory.select(RN)->nClusterings() == 0 )
      return false;

    // Now enfore merging scale cut if the event did not pass the merging scale
    // criterion.
    if ( nSteps > 0 && nSteps == nRequested && tmsnow < tmsval ) {
      infoPtr->errorMsg("Warning in VinciaMerging::cutOnProcess: Les Houches"
        " Event fails merging scale cut. Reject event.");
      return true;
    }

    bool complete = (FullHistory.select(RN)->nClusterings() == nSteps);

    if (!complete)
      infoPtr->errorMsg("Warning in Merging::cutOnProcess: No clusterings"
        " found. History incomplete.");

    if (debugNow) cout << " VinciaMerging::cutOnProcess end" << endl;

    // Done if only interested in cross section estimate after cuts.
    return false;

  }

  //--------------------------------------------------------------------------

  // Print statistics.

  void VinciaMerging::statistics() {
    // Recall switch to enfore merging scale cut.
    bool enforceCutOnLHE  = settingsPtr->flag("Merging:enforceCutOnLHE");
    // Recall merging scale value.
    double tmsval         = vcMergingHooksPtr->tms();
    bool printBanner      = enforceCutOnLHE && tmsNowMin > TMSMISMATCH*tmsval;
    // Reset minimal tms value.
    tmsNowMin             = infoPtr->eCM();
    if (!printBanner) return;
    // Header.
    cout << "\n *-------  VINCIA Matrix Element Merging Information  ------"
         << "-------------------------------------------------------*\n"
         << " |                                                            "
         << "                                                     |\n";
    // Print warning if the minimal tms value of any event was significantly
    // above the desired merging scale value.
    cout << " | Warning in Merging::statistics: All Les Houches events"
         << " significantly above Merging:TMS cut. Please check.       |\n";
    // Listing finished.
    cout << " |                                                            "
         << "                                                     |\n"
         << " *-------  End VINCIA Matrix Element Merging Information -----"
         << "-----------------------------------------------------*" << endl;
  }

}
